/*
** oalsound.cpp
** System interface for sound; uses OpenAL
**
**---------------------------------------------------------------------------
** Copyright 2008 Chris Robinson
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
**
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
** INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
** NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**---------------------------------------------------------------------------
**
*/

#include "doomstat.h"
#include "templates.h"
#include "oalsound.h"
#include "c_cvars.h"
#include "i_system.h"
#include "v_text.h"
#include "gi.h"
#include "actor.h"
#include "r_state.h"
#include "w_wad.h"
#include "i_music.h"
#include "i_musicinterns.h"
#include "tempfiles.h"


EXTERN_CVAR (Int, snd_channels)
EXTERN_CVAR (Int, snd_samplerate)
EXTERN_CVAR (Int, snd_driver)
EXTERN_CVAR (Int, snd_waterreverb)
CVAR (Bool, snd_efx, true, CVAR_ARCHIVE|CVAR_GLOBALCONFIG)
// Unfortunate enough, we can't use snd_waterlp because that specifies the
// cutoff frequency, whereas OpenAL wants the linear gain of the reference
// frequency (5khz).
CUSTOM_CVAR (Float, snd_waterhfgain, 0.1, CVAR_ARCHIVE|CVAR_GLOBALCONFIG)
{
	if (*self < 0.0f)
		self = 0.0f;
	else if (*self > 1.0f)
		self = 1.0f;
}


#ifndef NO_OPENAL

#ifndef _WIN32
#include <dlfcn.h>
#endif

#include <algorithm>
#include <memory>
#include <string>
#include <vector>

#define foreach(type, name, vec) \
	for(std::vector<type>::iterator (name) = (vec).begin(), \
	    (_end_##name) = (vec).end(); \
	    (name) != (_end_##name);(name)++)

#define MAKE_FUNC(x) typeof(::x)* OpenALSoundRenderer::x
// Standard OpenAL 1.1 functions
void *OpenALSoundRenderer::openal_handle;
MAKE_FUNC(alcCreateContext);
MAKE_FUNC(alcMakeContextCurrent);
MAKE_FUNC(alcProcessContext);
MAKE_FUNC(alcSuspendContext);
MAKE_FUNC(alcDestroyContext);
MAKE_FUNC(alcGetCurrentContext);
MAKE_FUNC(alcGetContextsDevice);
MAKE_FUNC(alcOpenDevice);
MAKE_FUNC(alcCloseDevice);
MAKE_FUNC(alcGetError);
MAKE_FUNC(alcIsExtensionPresent);
MAKE_FUNC(alcGetProcAddress);
MAKE_FUNC(alcGetEnumValue);
MAKE_FUNC(alcGetString);
MAKE_FUNC(alcGetIntegerv);
MAKE_FUNC(alcCaptureOpenDevice);
MAKE_FUNC(alcCaptureCloseDevice);
MAKE_FUNC(alcCaptureStart);
MAKE_FUNC(alcCaptureStop);
MAKE_FUNC(alcCaptureSamples);
MAKE_FUNC(alEnable);
MAKE_FUNC(alDisable);
MAKE_FUNC(alIsEnabled);
MAKE_FUNC(alGetString);
MAKE_FUNC(alGetBooleanv);
MAKE_FUNC(alGetIntegerv);
MAKE_FUNC(alGetFloatv);
MAKE_FUNC(alGetDoublev);
MAKE_FUNC(alGetBoolean);
MAKE_FUNC(alGetInteger);
MAKE_FUNC(alGetFloat);
MAKE_FUNC(alGetDouble);
MAKE_FUNC(alGetError);
MAKE_FUNC(alIsExtensionPresent);
MAKE_FUNC(alGetProcAddress);
MAKE_FUNC(alGetEnumValue);
MAKE_FUNC(alListenerf);
MAKE_FUNC(alListener3f);
MAKE_FUNC(alListenerfv);
MAKE_FUNC(alListeneri);
MAKE_FUNC(alGetListenerf);
MAKE_FUNC(alGetListener3f);
MAKE_FUNC(alGetListenerfv);
MAKE_FUNC(alGetListeneri);
MAKE_FUNC(alGetListeneriv);
MAKE_FUNC(alGenSources);
MAKE_FUNC(alDeleteSources);
MAKE_FUNC(alIsSource);
MAKE_FUNC(alSourcef);
MAKE_FUNC(alSource3f);
MAKE_FUNC(alSourcefv);
MAKE_FUNC(alSourcei);
MAKE_FUNC(alGetSourcef);
MAKE_FUNC(alGetSource3f);
MAKE_FUNC(alGetSourcefv);
MAKE_FUNC(alGetSourcei);
MAKE_FUNC(alGetSourceiv);
MAKE_FUNC(alSourcePlayv);
MAKE_FUNC(alSourceStopv);
MAKE_FUNC(alSourceRewindv);
MAKE_FUNC(alSourcePausev);
MAKE_FUNC(alSourcePlay);
MAKE_FUNC(alSourceStop);
MAKE_FUNC(alSourceRewind);
MAKE_FUNC(alSourcePause);
MAKE_FUNC(alSourceQueueBuffers);
MAKE_FUNC(alSourceUnqueueBuffers);
MAKE_FUNC(alGenBuffers);
MAKE_FUNC(alDeleteBuffers);
MAKE_FUNC(alIsBuffer);
MAKE_FUNC(alBufferData);
MAKE_FUNC(alGetBufferf);
MAKE_FUNC(alGetBufferfv);
MAKE_FUNC(alGetBufferi);
MAKE_FUNC(alGetBufferiv);
MAKE_FUNC(alDopplerFactor);
MAKE_FUNC(alDopplerVelocity);
MAKE_FUNC(alDistanceModel);
MAKE_FUNC(alListener3i);
MAKE_FUNC(alListeneriv);
MAKE_FUNC(alGetListener3i);
MAKE_FUNC(alSource3i);
MAKE_FUNC(alSourceiv);
MAKE_FUNC(alGetSource3i);
MAKE_FUNC(alBufferf);
MAKE_FUNC(alBuffer3f);
MAKE_FUNC(alBufferfv);
MAKE_FUNC(alBufferi);
MAKE_FUNC(alBuffer3i);
MAKE_FUNC(alBufferiv);
MAKE_FUNC(alGetBuffer3f);
MAKE_FUNC(alGetBuffer3i);
MAKE_FUNC(alSpeedOfSound);
// ALURE 1.0 functions
void *OpenALSoundRenderer::alure_handle;
MAKE_FUNC(alureGetErrorString);
MAKE_FUNC(alureGetVersion);
MAKE_FUNC(alureGetDeviceNames);
MAKE_FUNC(alureFreeDeviceNames);
MAKE_FUNC(alureInitDevice);
MAKE_FUNC(alureShutdownDevice);
MAKE_FUNC(alureGetSampleFormat);
MAKE_FUNC(alureSleep);
MAKE_FUNC(alureCreateBufferFromFile);
MAKE_FUNC(alureCreateBufferFromMemory);
MAKE_FUNC(alureBufferDataFromFile);
MAKE_FUNC(alureBufferDataFromMemory);
MAKE_FUNC(alureCreateStreamFromFile);
MAKE_FUNC(alureCreateStreamFromMemory);
MAKE_FUNC(alureCreateStreamFromStaticMemory);
MAKE_FUNC(alureCreateStreamFromCallback);
MAKE_FUNC(alureBufferDataFromStream);
MAKE_FUNC(alureRewindStream);
MAKE_FUNC(alureDestroyStream);
MAKE_FUNC(alureInstallDecodeCallbacks);
#undef MAKE_FUNC

extern ReverbContainer *ForcedEnvironment;

#define PITCH_MULT (0.7937005f) /* Approx. 4 semitones lower; what Nash suggested */


static float GetRolloff(const FRolloffInfo *rolloff, float distance)
{
	if(distance <= rolloff->MinDistance)
		return 1;
	// Logarithmic rolloff has no max distance where it goes silent.
	if(rolloff->RolloffType == ROLLOFF_Log)
		return rolloff->MinDistance / (rolloff->MinDistance + rolloff->RolloffFactor * (distance - rolloff->MinDistance));
	if(distance >= rolloff->MaxDistance)
		return 0;

	float volume = (rolloff->MaxDistance - distance) / (rolloff->MaxDistance - rolloff->MinDistance);
	if(rolloff->RolloffType == ROLLOFF_Custom && S_SoundCurve != NULL)
		return S_SoundCurve[int(S_SoundCurveSize * (1 - volume))] / 127.f;
//	if(rolloff->RolloffType == ROLLOFF_Doom)
//		return volume * (rolloff->MinDistance/distance);
	return volume;
}


class OpenALSoundStream : public SoundStream
{
	OpenALSoundRenderer *Renderer;
	alureStream *Stream;
	ALuint Source;

	static const ALuint ChunkLen = 32*1024;
	static const ALuint NumBuffers = 3;
	ALuint Buffers[NumBuffers];
	bool Playing;
	bool Looping;
	bool Done;

	ALuint BufferLen;
	ALuint Freq;
	unsigned int BaseLen;

	virtual bool SetupSource()
	{
		if(Renderer->FreeSfx.size() == 0)
		{
			FSoundChan *schan = Channels;
			FSoundChan *lowest = NULL;
			while(schan)
			{
				if(schan->SysChannel != NULL)
				{
					if(!lowest || schan->Priority < lowest->Priority ||
					   (schan->Priority == lowest->Priority &&
					    schan->Distance >= lowest->Distance))
						lowest = schan;
				}
				schan = schan->NextChan;
			}
			if(lowest)
				Renderer->StopChannel(lowest);
			if(Renderer->FreeSfx.size() == 0)
				return false;
		}
		Source = Renderer->FreeSfx.back();
		Renderer->FreeSfx.pop_back();

		Renderer->alSource3f(Source, AL_DIRECTION, 0.0f, 0.0f, 0.0f);
		Renderer->alSource3f(Source, AL_VELOCITY, 0.0f, 0.0f, 0.0f);
		Renderer->alSource3f(Source, AL_POSITION, 0.0f, 0.0f, 0.0f);
		Renderer->alSourcef(Source, AL_ROLLOFF_FACTOR, 0.0f);
		Renderer->alSourcef(Source, AL_PITCH, 1.0f);
		Renderer->alSourcef(Source, AL_GAIN, Volume*Renderer->MusicVolume);
		if(Renderer->EnvSlot)
			Renderer->alSource3i(Source, AL_AUXILIARY_SEND_FILTER, 0, 0, AL_FILTER_NULL);
		if(Renderer->WaterLP)
			Renderer->alSourcei(Source, AL_DIRECT_FILTER, AL_FILTER_NULL);
		Renderer->alSourcei(Source, AL_SOURCE_RELATIVE, AL_TRUE);
		Renderer->alSourcei(Source, AL_LOOPING, AL_FALSE);

		return (Renderer->alGetError() == AL_NO_ERROR);
	}

	SoundStreamCallback Callback;
	void *UserData;
	static ALuint WrapCallback(void *userdata, ALubyte *data, ALuint bytes)
	{
		OpenALSoundStream *This = (OpenALSoundStream*)userdata;
		if(This->Done)
			return 0;

		if(!This->Callback(This, data, bytes, This->UserData))
		{
			This->Done = true;
			return 0;
		}
		return bytes;
	}

public:
	FTempFileName tmpfile;
	ALfloat Volume;

	OpenALSoundStream(OpenALSoundRenderer *renderer)
		: Renderer(renderer), Stream(NULL), Source(0), Playing(false),
		  Looping(false), Done(true), Volume(1.0f)
	{
		memset(Buffers, 0, sizeof(Buffers));
		Renderer->Streams.push_back(this);
	}

	virtual ~OpenALSoundStream()
	{
		if(Source)
		{
			Renderer->alSourceStop(Source);
			Renderer->alSourcei(Source, AL_BUFFER, 0);
			Renderer->FreeSfx.push_back(Source);
			Source = 0;
		}

		if(!Renderer->alureDestroyStream(Stream, NumBuffers, Buffers))
			Printf("Destroy stream error: %s\n", Renderer->alureGetErrorString());
		Stream = NULL;
		memset(Buffers, 0, sizeof(Buffers));

		Renderer->Streams.erase(find(Renderer->Streams.begin(),
		                             Renderer->Streams.end(), this));
		Renderer = NULL;
	}

	virtual bool Play(bool looping, float vol)
	{
		if(!Done)
			return true;

		BaseLen = 0;
		Playing = false;
		Looping = looping;
		SetVolume(vol);

		ALuint i = 0;
		ALenum err;

		Renderer->alSourcei(Source, AL_BUFFER, 0);
		Renderer->alSourceQueueBuffers(Source, NumBuffers, Buffers);
		if((err=Renderer->alGetError()) == AL_NO_ERROR)
		{
			Renderer->alSourcePlay(Source);
			if((err=Renderer->alGetError()) == AL_NO_ERROR)
			{
				Done = false;
				Playing = true;
				return true;
			}
		}
		return false;
	}

	virtual void Stop()
	{
		Playing = false;
		Done = true;
		BaseLen = 0;
		Renderer->alSourceStop(Source);
		Renderer->alGetError();
	}

	virtual bool SetPaused(bool paused)
    {
		if(Done)
			return false;

		if(paused)
		{
			Renderer->alSourcePause(Source);
			if(Renderer->alGetError() != AL_NO_ERROR)
				return false;
		}
		// The next Poll() call will automatically unpause it as needed, after
		// filling any used buffers
		Playing = !paused;
		return true;
	}

	virtual void SetVolume(float vol)
	{
		Volume = vol;
		Renderer->alSourcef(Source, AL_GAIN, Volume*Renderer->MusicVolume);
		Renderer->alGetError();
	}

	virtual unsigned int GetPosition()
	{
		ALfloat pos;
		Renderer->alGetSourcef(Source, AL_SEC_OFFSET, &pos);
		if(Renderer->alGetError() != AL_NO_ERROR)
			return (unsigned int)(pos*1000.0f) + (BaseLen*1000/Freq);
		return 0;
	}

	virtual bool IsEnded()
	{
		return Done;
	}

	virtual bool Poll()
	{
		if(!Playing || Done)
			return true;

		ALuint bufs[NumBuffers];
		ALint processed;
		ALint state;
		ALenum err;

		Renderer->alGetSourcei(Source, AL_SOURCE_STATE, &state);
		Renderer->alGetSourcei(Source, AL_BUFFERS_PROCESSED, &processed);
		if((err=Renderer->alGetError()) != AL_NO_ERROR)
		{
			Done = true;
			return false;
		}

		Renderer->alSourceUnqueueBuffers(Source, processed, bufs);
		if((err=Renderer->alGetError()) != AL_NO_ERROR)
		{
			Done = true;
			return false;
		}
		BaseLen += BufferLen*processed;

		int filled = Renderer->alureBufferDataFromStream(Stream, processed, bufs);
		if(filled < 0)
		{
			Done = true;
			return false;
		}
		while(filled < processed && Looping)
		{
			if(!Renderer->alureRewindStream(Stream))
				break;

			ALsizei got;
			got = Renderer->alureBufferDataFromStream(Stream, processed-filled, &bufs[filled]);
			if(got <= 0)
				break;
			filled += got;
		}

		if(filled < processed)
			Done = true;

		Renderer->alSourceQueueBuffers(Source, filled, bufs);
		if((err=Renderer->alGetError()) != AL_NO_ERROR)
		{
			Done = true;
			return false;
		}

		if(state != AL_PLAYING)
		{
			Renderer->alSourcePlay(Source);
			if((err=Renderer->alGetError()) != AL_NO_ERROR)
			{
				Done = true;
				return false;
			}
		}
		return true;
	}


	virtual bool Init(const char *filename)
	{
		Stream = Renderer->alureCreateStreamFromFile(filename, ChunkLen, NumBuffers, Buffers);
		if(!Stream)
		{
			Printf("Failed to create stream: %s\n", Renderer->alureGetErrorString());
			return false;
		}

		ALint bits, channels, size;
		Renderer->alGetBufferi(Buffers[0], AL_FREQUENCY, (ALint*)&Freq);
		Renderer->alGetBufferi(Buffers[0], AL_BITS, &bits);
		Renderer->alGetBufferi(Buffers[0], AL_CHANNELS, &channels);
		Renderer->alGetBufferi(Buffers[0], AL_SIZE, &size);
		if(Renderer->alGetError() != AL_NO_ERROR)
			return false;

		BufferLen = size / channels * 8 / bits;
		return SetupSource();
	}

	virtual bool Init(const BYTE *data, unsigned int datalen)
	{
		// FIXME: Use alureCreateStreamFromStaticMemory?
		Stream = Renderer->alureCreateStreamFromMemory(data, datalen, ChunkLen, NumBuffers, Buffers);
		if(!Stream)
		{
			Printf("Failed to create stream: %s\n", Renderer->alureGetErrorString());
			return false;
		}

		ALint bits, channels, size;
		Renderer->alGetBufferi(Buffers[0], AL_FREQUENCY, (ALint*)&Freq);
		Renderer->alGetBufferi(Buffers[0], AL_BITS, &bits);
		Renderer->alGetBufferi(Buffers[0], AL_CHANNELS, &channels);
		Renderer->alGetBufferi(Buffers[0], AL_SIZE, &size);
		if(Renderer->alGetError() != AL_NO_ERROR)
			return false;

		BufferLen = size / channels * 8 / bits;
		return SetupSource();
	}

	virtual bool Init(int bits, int channels, int freq, int buffsize, SoundStreamCallback cb, void *userdata)
	{
		Freq = freq;
		Callback = cb;
		UserData = userdata;

		int blockSize = bits * channels / 8;
		ALenum format;
		if(bits == 32) // 32-bit is always float
			format = Renderer->alureGetSampleFormat(channels, 0, bits);
		else
			format = Renderer->alureGetSampleFormat(channels, bits, 0);

		if(format == AL_NONE)
		{
			Printf("Failed to get format for %d-bit, %d channels: %s\n", bits, channels, Renderer->alureGetErrorString());
			return false;
		}

		BufferLen = buffsize / blockSize;
		Stream = Renderer->alureCreateStreamFromCallback(WrapCallback, this, format,
		                                                 Freq, BufferLen*blockSize,
		                                                 NumBuffers, Buffers);
		if(!Stream)
		{
			Printf("Failed to create stream: %s\n", Renderer->alureGetErrorString());
			return false;
		}

		return SetupSource();
	}
};


OpenALSoundRenderer::OpenALSoundRenderer()
	: Context(NULL), SrcDistanceModel(false), SFXPaused(0),
	  PrevEnvironment(NULL), EnvSlot(0), WaterLP(0)
{
	Printf ("I_InitSound: Initializing OpenAL\n");

#ifdef _WIN32
	if(!openal_handle)
	{
		openal_handle = LoadLibrary("OpenAL32.dll");
		if(openal_handle == NULL)
		{
			Printf("Failed to load OpenAL32.dll\n");
			return;
		}
		DPrintf("OpenAL32.dll loaded @ %p\n", openal_handle);
	}
	if(!alure_handle)
	{
		alure_handle = LoadLibrary("ALURE32.dll");
		if(alure_handle == NULL)
		{
			Printf("Failed to load ALURE32.dll\n");
			return;
		}
		DPrintf("ALURE32.dll loaded @ %p\n", alure_handle);
	}

#define LOAD_FUNC(m, x) do { \
	x = reinterpret_cast<typeof(x)>(GetProcAddress((HMODULE)m##_handle, #x)); \
	if(x == NULL) { \
		Printf("Failed to load function "#x"from "#m"\n"); \
		FreeLibrary(m##_handle); \
		m##_handle = NULL; \
		return; \
	} \
} while(0)

#elif defined(__APPLE__)

	const char *errstr;
	openal_handle = RTLD_DEFAULT; // Already linked
	if(!alure_handle)
	{
		alure_handle = dlopen("libalure.1.dylib", RTLD_NOW);
		if((errstr=dlerror()) != NULL)
		{
			Printf("Failed to load libalure.1.dylib: %s\n", errstr);
			return;
		}
		DPrintf("libalure.1.dylib loaded @ %p\n", alure_handle);
	}

#define LOAD_FUNC(m, x) do { \
	x = reinterpret_cast<typeof(x)>(dlsym(m##_handle, #x)); \
	if((errstr=dlerror()) != NULL) { \
		Printf("Failed to load function "#x"from "#m": %s\n", errstr); \
		if(m##_handle != RTLD_DEFAULT) \
			dlclose(m##_handle); \
		m##_handle = NULL; \
		return; \
	} \
} while(0)

#else

	const char *errstr;
	if(!openal_handle)
	{
		openal_handle = dlopen("libopenal.so.1", RTLD_NOW);
		if((errstr=dlerror()) != NULL)
		{
			Printf("Failed to load libopenal.so.1: %s\n", errstr);
			return;
		}
		DPrintf("libopenal.so.1 loaded @ %p\n", openal_handle);
	}
	if(!alure_handle)
	{
		alure_handle = dlopen("libalure.so.1", RTLD_NOW);
		if((errstr=dlerror()) != NULL)
		{
			Printf("Failed to load libalure.so.1: %s\n", errstr);
			return;
		}
		DPrintf("libalure.so.1 loaded @ %p\n", alure_handle);
	}

#define LOAD_FUNC(m, x) do { \
	x = reinterpret_cast<typeof(x)>(dlsym(m##_handle, #x)); \
	if((errstr=dlerror()) != NULL) { \
		Printf("Failed to load function "#x"from "#m": %s\n", errstr); \
		dlclose(m##_handle); \
		m##_handle = NULL; \
		return; \
	} \
} while(0)
#endif

	// OpenAL 1.1 functions
	LOAD_FUNC(openal, alcCreateContext);
	LOAD_FUNC(openal, alcMakeContextCurrent);
	LOAD_FUNC(openal, alcProcessContext);
	LOAD_FUNC(openal, alcSuspendContext);
	LOAD_FUNC(openal, alcDestroyContext);
	LOAD_FUNC(openal, alcGetCurrentContext);
	LOAD_FUNC(openal, alcGetContextsDevice);
	LOAD_FUNC(openal, alcOpenDevice);
	LOAD_FUNC(openal, alcCloseDevice);
	LOAD_FUNC(openal, alcGetError);
	LOAD_FUNC(openal, alcIsExtensionPresent);
	LOAD_FUNC(openal, alcGetProcAddress);
	LOAD_FUNC(openal, alcGetEnumValue);
	LOAD_FUNC(openal, alcGetString);
	LOAD_FUNC(openal, alcGetIntegerv);
	LOAD_FUNC(openal, alcCaptureOpenDevice);
	LOAD_FUNC(openal, alcCaptureCloseDevice);
	LOAD_FUNC(openal, alcCaptureStart);
	LOAD_FUNC(openal, alcCaptureStop);
	LOAD_FUNC(openal, alcCaptureSamples);
	LOAD_FUNC(openal, alEnable);
	LOAD_FUNC(openal, alDisable);
	LOAD_FUNC(openal, alIsEnabled);
	LOAD_FUNC(openal, alGetString);
	LOAD_FUNC(openal, alGetBooleanv);
	LOAD_FUNC(openal, alGetIntegerv);
	LOAD_FUNC(openal, alGetFloatv);
	LOAD_FUNC(openal, alGetDoublev);
	LOAD_FUNC(openal, alGetBoolean);
	LOAD_FUNC(openal, alGetInteger);
	LOAD_FUNC(openal, alGetFloat);
	LOAD_FUNC(openal, alGetDouble);
	LOAD_FUNC(openal, alGetError);
	LOAD_FUNC(openal, alIsExtensionPresent);
	LOAD_FUNC(openal, alGetProcAddress);
	LOAD_FUNC(openal, alGetEnumValue);
	LOAD_FUNC(openal, alListenerf);
	LOAD_FUNC(openal, alListener3f);
	LOAD_FUNC(openal, alListenerfv);
	LOAD_FUNC(openal, alListeneri);
	LOAD_FUNC(openal, alGetListenerf);
	LOAD_FUNC(openal, alGetListener3f);
	LOAD_FUNC(openal, alGetListenerfv);
	LOAD_FUNC(openal, alGetListeneri);
	LOAD_FUNC(openal, alGetListeneriv);
	LOAD_FUNC(openal, alGenSources);
	LOAD_FUNC(openal, alDeleteSources);
	LOAD_FUNC(openal, alIsSource);
	LOAD_FUNC(openal, alSourcef);
	LOAD_FUNC(openal, alSource3f);
	LOAD_FUNC(openal, alSourcefv);
	LOAD_FUNC(openal, alSourcei);
	LOAD_FUNC(openal, alGetSourcef);
	LOAD_FUNC(openal, alGetSource3f);
	LOAD_FUNC(openal, alGetSourcefv);
	LOAD_FUNC(openal, alGetSourcei);
	LOAD_FUNC(openal, alGetSourceiv);
	LOAD_FUNC(openal, alSourcePlayv);
	LOAD_FUNC(openal, alSourceStopv);
	LOAD_FUNC(openal, alSourceRewindv);
	LOAD_FUNC(openal, alSourcePausev);
	LOAD_FUNC(openal, alSourcePlay);
	LOAD_FUNC(openal, alSourceStop);
	LOAD_FUNC(openal, alSourceRewind);
	LOAD_FUNC(openal, alSourcePause);
	LOAD_FUNC(openal, alSourceQueueBuffers);
	LOAD_FUNC(openal, alSourceUnqueueBuffers);
	LOAD_FUNC(openal, alGenBuffers);
	LOAD_FUNC(openal, alDeleteBuffers);
	LOAD_FUNC(openal, alIsBuffer);
	LOAD_FUNC(openal, alBufferData);
	LOAD_FUNC(openal, alGetBufferf);
	LOAD_FUNC(openal, alGetBufferfv);
	LOAD_FUNC(openal, alGetBufferi);
	LOAD_FUNC(openal, alGetBufferiv);
	LOAD_FUNC(openal, alDopplerFactor);
	LOAD_FUNC(openal, alDopplerVelocity);
	LOAD_FUNC(openal, alDistanceModel);
	LOAD_FUNC(openal, alListener3i);
	LOAD_FUNC(openal, alListeneriv);
	LOAD_FUNC(openal, alGetListener3i);
	LOAD_FUNC(openal, alSource3i);
	LOAD_FUNC(openal, alSourceiv);
	LOAD_FUNC(openal, alGetSource3i);
	LOAD_FUNC(openal, alBufferf);
	LOAD_FUNC(openal, alBuffer3f);
	LOAD_FUNC(openal, alBufferfv);
	LOAD_FUNC(openal, alBufferi);
	LOAD_FUNC(openal, alBuffer3i);
	LOAD_FUNC(openal, alBufferiv);
	LOAD_FUNC(openal, alGetBuffer3f);
	LOAD_FUNC(openal, alGetBuffer3i);
	LOAD_FUNC(openal, alSpeedOfSound);
	// ALURE 1.0 functions
	LOAD_FUNC(alure, alureGetErrorString);
	LOAD_FUNC(alure, alureGetVersion);
	LOAD_FUNC(alure, alureGetDeviceNames);
	LOAD_FUNC(alure, alureFreeDeviceNames);
	LOAD_FUNC(alure, alureInitDevice);
	LOAD_FUNC(alure, alureShutdownDevice);
	LOAD_FUNC(alure, alureGetSampleFormat);
	LOAD_FUNC(alure, alureSleep);
	LOAD_FUNC(alure, alureCreateBufferFromFile);
	LOAD_FUNC(alure, alureCreateBufferFromMemory);
	LOAD_FUNC(alure, alureBufferDataFromFile);
	LOAD_FUNC(alure, alureBufferDataFromMemory);
	LOAD_FUNC(alure, alureCreateStreamFromFile);
	LOAD_FUNC(alure, alureCreateStreamFromMemory);
	LOAD_FUNC(alure, alureCreateStreamFromStaticMemory);
	LOAD_FUNC(alure, alureCreateStreamFromCallback);
	LOAD_FUNC(alure, alureBufferDataFromStream);
	LOAD_FUNC(alure, alureRewindStream);
	LOAD_FUNC(alure, alureDestroyStream);
	LOAD_FUNC(alure, alureInstallDecodeCallbacks);
#undef LOAD_FUNC

	ALuint major, minor;
	alureGetVersion(&major, &minor);
	Printf(" Using ALURE %u.%u\n", major, minor);

	std::vector<ALCint> attribs;
	attribs.push_back(ALC_FREQUENCY);
	attribs.push_back(*snd_samplerate);
	// Other attribs..?

	attribs.push_back(0);

	ALCsizei devcount = 0;
	const ALCchar **devnames = alureGetDeviceNames(ALC_TRUE, &devcount);
	if(devnames)
	{
		int idx = *snd_driver;
		if(idx < devcount)
		{
			if(alureInitDevice(devnames[idx], &attribs[0]) == AL_FALSE)
				Printf(TEXTCOLOR_BLUE" Failed to open driver %s (%s). Trying default.\n", devnames[idx], alureGetErrorString());
			else
				Context = alcGetCurrentContext();
		}
		else
			Printf(TEXTCOLOR_BLUE" Driver %d does not exist. Using default.\n", idx);

		alureFreeDeviceNames(devnames);
		devnames = NULL;
	}

	if(!Context)
	{
		if(alureInitDevice(NULL, &attribs[0]) == AL_FALSE)
		{
			Printf(TEXTCOLOR_RED" Could not open device: %s\n", alureGetErrorString());
			return;
		}

		Context = alcGetCurrentContext();
	}

	ALCdevice *Device = alcGetContextsDevice(Context);
	Printf("  Opened device %s\n", alcGetString(Device, ALC_DEVICE_SPECIFIER));

	if(alcIsExtensionPresent(Device, "ALC_EXT_EFX"))
	{
		ALCint major, minor, sends;
		alcGetIntegerv(Device, ALC_EFX_MAJOR_VERSION, 1, &major);
		alcGetIntegerv(Device, ALC_EFX_MINOR_VERSION, 1, &minor);
		alcGetIntegerv(Device, ALC_MAX_AUXILIARY_SENDS, 1, &sends);
		if(alcGetError(Device) == AL_NO_ERROR)
			Printf("   EFX %d.%d (%d aux send%s) supported\n", major, minor, sends, (sends==1)?"":"s");
	}

	if((SrcDistanceModel=alIsExtensionPresent("AL_EXTX_source_distance_model")))
		Printf("   Per-source distance models supported\n");
	if(alIsExtensionPresent("AL_EXT_FLOAT32"))
		Printf("   Floating-point formats supported\n");
	if(alIsExtensionPresent("AL_EXT_MCFORMATS"))
		Printf("   Multi-channel formats supported\n");

	alDopplerFactor(0.5f);
	// Equivilant to setting the deprecated alDopplerVelocity to 96.0f
	alSpeedOfSound(343.3f * 96.0f);

	ALenum err = alGetError();
	if(err != AL_NO_ERROR)
	{
		Printf(TEXTCOLOR_RED"OpenAL error: %s\n", alGetString(err));
		alureShutdownDevice();
		Context = NULL;
		return;
	}

	Sources.resize(std::max(*snd_channels, 16));
	for(size_t i = 0;i < Sources.size();i++)
	{
		alGenSources(1, &Sources[i]);
		if(alGetError() != AL_NO_ERROR)
		{
			Sources.resize(i);
			break;
		}
		FreeSfx.push_back(Sources[i]);
	}

	if(Sources.size() == 0)
	{
		Printf(TEXTCOLOR_RED"OpenAL error: could not generate any sound sources!\n");
		alureShutdownDevice();
		Context = NULL;
		return;
	}

	LastWaterLP = 1.0f;
	if(*snd_efx && alcIsExtensionPresent(Device, "ALC_EXT_EFX"))
	{
		// EFX function pointers
#define LOAD_FUNC(x)  (x = reinterpret_cast<typeof(x)>(alGetProcAddress(#x)))
		LOAD_FUNC(alGenEffects);
		LOAD_FUNC(alDeleteEffects);
		LOAD_FUNC(alIsEffect);
		LOAD_FUNC(alEffecti);
		LOAD_FUNC(alEffectiv);
		LOAD_FUNC(alEffectf);
		LOAD_FUNC(alEffectfv);
		LOAD_FUNC(alGetEffecti);
		LOAD_FUNC(alGetEffectiv);
		LOAD_FUNC(alGetEffectf);
		LOAD_FUNC(alGetEffectfv);

		LOAD_FUNC(alGenFilters);
		LOAD_FUNC(alDeleteFilters);
		LOAD_FUNC(alIsFilter);
		LOAD_FUNC(alFilteri);
		LOAD_FUNC(alFilteriv);
		LOAD_FUNC(alFilterf);
		LOAD_FUNC(alFilterfv);
		LOAD_FUNC(alGetFilteri);
		LOAD_FUNC(alGetFilteriv);
		LOAD_FUNC(alGetFilterf);
		LOAD_FUNC(alGetFilterfv);

		LOAD_FUNC(alGenAuxiliaryEffectSlots);
		LOAD_FUNC(alDeleteAuxiliaryEffectSlots);
		LOAD_FUNC(alIsAuxiliaryEffectSlot);
		LOAD_FUNC(alAuxiliaryEffectSloti);
		LOAD_FUNC(alAuxiliaryEffectSlotiv);
		LOAD_FUNC(alAuxiliaryEffectSlotf);
		LOAD_FUNC(alAuxiliaryEffectSlotfv);
		LOAD_FUNC(alGetAuxiliaryEffectSloti);
		LOAD_FUNC(alGetAuxiliaryEffectSlotiv);
		LOAD_FUNC(alGetAuxiliaryEffectSlotf);
		LOAD_FUNC(alGetAuxiliaryEffectSlotfv);
#undef LOAD_FUNC
		if(alGetError() == AL_NO_ERROR)
		{
			alGenAuxiliaryEffectSlots(1, &EnvSlot);
			alGetError();

			alGenFilters(1, &WaterLP);
			if(alGetError() == AL_NO_ERROR)
			{
				alFilteri(WaterLP, AL_FILTER_TYPE, AL_FILTER_LOWPASS);
				if(alGetError() != AL_NO_ERROR)
				{
					alDeleteFilters(1, &WaterLP);
					WaterLP = 0;
					alGetError();
				}
			}
		}
	}
}

OpenALSoundRenderer::~OpenALSoundRenderer ()
{
	if(!Context)
		return;

	foreach(OpenALSoundStream*, i, Streams)
		delete *i;
	Streams.clear();

	if(WaterLP)
		alDeleteFilters(1, &WaterLP);
	WaterLP = 0;

	for(EffectMap::iterator i = EnvEffects.begin();i != EnvEffects.end();i++)
		alDeleteEffects(1, &(i->second));
	EnvEffects.clear();

	alDeleteSources(Sources.size(), &Sources[0]);
	Sources.clear();
	FreeSfx.clear();
	SfxGroup.clear();
	PausableSfx.clear();

	if(EnvSlot)
		alDeleteAuxiliaryEffectSlots(1, &EnvSlot);
	EnvSlot = 0;

	alureShutdownDevice();
	Context = NULL;
}

void OpenALSoundRenderer::SetSfxVolume(float volume)
{
	SfxVolume = volume;

	FSoundChan *schan = Channels;
	while(schan)
	{
		if(schan->SysChannel != NULL)
		{
			ALuint source = *static_cast<ALuint*>(schan->SysChannel);
			if(schan->ManualGain)
			{
				float gain = GetRolloff(&schan->Rolloff, schan->Distance);
				alSourcef(source, AL_GAIN, volume*gain);
			}
			else
				alSourcef(source, AL_GAIN, volume);
		}
		schan = schan->NextChan;
	}

	alGetError();
}

void OpenALSoundRenderer::SetMusicVolume(float volume)
{
	MusicVolume = volume;
	foreach(OpenALSoundStream*, i, Streams)
		(*i)->SetVolume((*i)->Volume);
}

unsigned int OpenALSoundRenderer::GetMSLength(SoundHandle sfx)
{
	if(sfx.data)
	{
		ALuint buffer = *static_cast<ALuint*>(sfx.data);
		ALint bits, channels, freq, size;
		alGetBufferi(buffer, AL_BITS, &bits);
		alGetBufferi(buffer, AL_CHANNELS, &channels);
		alGetBufferi(buffer, AL_FREQUENCY, &freq);
		alGetBufferi(buffer, AL_SIZE, &size);
		if(alGetError() == AL_NO_ERROR)
			return size / channels * 8 / bits * 1000 / freq;
	}
	return 0;
}

unsigned int OpenALSoundRenderer::GetSampleLength(SoundHandle sfx)
{
	if(sfx.data)
	{
		ALuint buffer = *static_cast<ALuint*>(sfx.data);
		ALint bits, channels, size;
		alGetBufferi(buffer, AL_BITS, &bits);
		alGetBufferi(buffer, AL_CHANNELS, &channels);
		alGetBufferi(buffer, AL_SIZE, &size);
		if(alGetError() == AL_NO_ERROR)
			return size / channels * 8 / bits;
	}
	return 0;
}

float OpenALSoundRenderer::GetOutputRate()
{
	ALCdevice *Device = alcGetContextsDevice(Context);
	ALint rate = *snd_samplerate; // Default, just in case

	alcGetIntegerv(Device, ALC_FREQUENCY, 1, &rate);
	alcGetError(Device);

	return (float)rate;
}


SoundHandle OpenALSoundRenderer::LoadSoundRaw(BYTE *sfxdata, int length, int frequency, int channels, int bits)
{
	ALenum err;
	ALuint buffer;
	SoundHandle retval = { NULL };

	if(length == 0) return retval;

	if(bits == -8)
	{
		// Convert signed to unsigned
		for (int i = 0; i < length; ++i)
			sfxdata[i] ^= 0x80;
		bits = -bits;
	}

	// Assume Raw sounds are always integer samples, not float
	ALenum format = alureGetSampleFormat(channels, bits, 0);
	if(format == AL_NONE)
	{
		Printf("Failed to get format for %d-bit, %d channels: %s\n", bits, channels, alureGetErrorString());
		return retval;
	}

	alGenBuffers(1, &buffer);
	if((err=alGetError()) != AL_NO_ERROR)
	{
		DPrintf("Failed to allocate buffer: %s\n", alGetString(err));
		return retval;
	}

	length -= length%(channels*bits/8);
	alBufferData(buffer, format, sfxdata, length, frequency);
	if((err=alGetError()) != AL_NO_ERROR)
	{
		DPrintf("Failed to buffer data: %s\n", alGetString(err));
		alDeleteBuffers(1, &buffer);
		alGetError();
		return retval;
	}

	retval.data = new ALuint(buffer);
	return retval;
}

SoundHandle OpenALSoundRenderer::LoadSound(BYTE *sfxdata, int length)
{
	SoundHandle retval = { NULL };
	ALuint buffer;

	buffer = alureCreateBufferFromMemory(sfxdata, length);
	if(buffer == AL_NONE)
	{
		DPrintf("Failed to load sound: %s\n", alureGetErrorString());
		return retval;
	}

	retval.data = new ALuint(buffer);
	return retval;
}

void OpenALSoundRenderer::UnloadSound(SoundHandle sfx)
{
	if(!sfx.data)
		return;
	ALuint buffer = *static_cast<ALuint*>(sfx.data);

	delete static_cast<ALuint*>(sfx.data);
	sfx.data = NULL;

	// Stop any channels using this buffer or else it can't be deleted
	FSoundChan *schan = Channels;
	while(schan)
	{
		if(schan->SysChannel != NULL)
		{
			ALint curbuf;
			alGetSourcei(*static_cast<ALuint*>(schan->SysChannel), AL_BUFFER, &curbuf);
			if(alGetError() != AL_NO_ERROR || static_cast<ALuint>(curbuf) == buffer)
				StopChannel(schan);
		}
		schan = schan->NextChan;
	}

	alDeleteBuffers(1, &buffer);
	alGetError();
}

SoundStream *OpenALSoundRenderer::CreateStream(SoundStreamCallback callback, int buffbytes, int flags, int samplerate, void *userdata)
{
	int bits = ((flags&SoundStream::Bits8) ? 8 :
	            ((flags&SoundStream::Bits32) ? 0 :
	             ((flags&SoundStream::Float) ? 32 : 16)));
	int channels = ((flags&SoundStream::Mono) ? 1 : 2);

	std::auto_ptr<OpenALSoundStream> stream(new OpenALSoundStream(this));
	if(!stream->Init(bits, channels, samplerate, buffbytes, callback, userdata))
		return NULL;

	return stream.release();
}

SoundStream *OpenALSoundRenderer::OpenStream(const char *filename, int flags, int offset, int length)
{
	std::auto_ptr<OpenALSoundStream> stream(new OpenALSoundStream(this));

	if(offset > 0)
	{
		// If there's an offset to the start of the data, seperate it into its
		// own temp file
		FILE *infile = fopen(filename, "rb");
		FILE *f = fopen(stream->tmpfile, "wb");
		if(!infile || !f || fseek(infile, offset, SEEK_SET) != 0)
		{
			if(infile) fclose(infile);
			if(f) fclose(f);
			return NULL;
		}
		BYTE buf[1024];
		size_t got;
		do {
			got = std::min(sizeof(buf), (size_t)length);
			got = fread(buf, 1, got, infile);
			if(got == 0)
				break;
		} while(fwrite(buf, 1, got, f) == got && (length-=got) > 0);
		fclose(f);
		fclose(infile);

		filename = stream->tmpfile;
	}

	bool ok = ((offset == -1) ? stream->Init((const BYTE*)filename, length) :
	                            stream->Init(filename));
	if(ok == false)
		return NULL;

	return stream.release();
}

FISoundChannel *OpenALSoundRenderer::StartSound(SoundHandle sfx, float vol, int pitch, int chanflags, FISoundChannel *reuse_chan)
{
	if(FreeSfx.size() == 0)
	{
		FSoundChan *schan = Channels;
		FSoundChan *lowest = NULL;
		while(schan)
		{
			if(schan->SysChannel != NULL)
			{
				if(!lowest || schan->Priority < lowest->Priority ||
				   (schan->Priority == lowest->Priority &&
				    schan->Distance > lowest->Distance))
					lowest = schan;
			}
			schan = schan->NextChan;
		}
		if(lowest)
			StopChannel(lowest);

		if(FreeSfx.size() == 0)
			return NULL;
	}

	ALuint &source = *find(Sources.begin(), Sources.end(), FreeSfx.back());
	alSource3f(source, AL_DIRECTION, 0.0f, 0.0f, 0.0f);
	alSource3f(source, AL_VELOCITY, 0.0f, 0.0f, 0.0f);
	alSource3f(source, AL_POSITION, 0.0f, 0.0f, 0.0f);
	alSourcei(source, AL_SOURCE_RELATIVE, AL_TRUE);
	alSourcei(source, AL_LOOPING, (chanflags&SNDF_LOOP)?AL_TRUE:AL_FALSE);
	alSourcef(source, AL_ROLLOFF_FACTOR, 0.0f);
	alSourcef(source, AL_GAIN, SfxVolume*vol);
	if(LastWaterLP < 1.0f)
	{
		alSourcef(source, AL_PITCH, pitch/128.0f*PITCH_MULT);
		if(WaterLP)
			alSourcei(source, AL_DIRECT_FILTER, WaterLP);
	}
	else
	{
		alSourcef(source, AL_PITCH, pitch/128.0f);
		if(WaterLP)
			alSourcei(source, AL_DIRECT_FILTER, AL_FILTER_NULL);
	}
	if(EnvSlot)
	{
		alSourcef(source, AL_ROOM_ROLLOFF_FACTOR, 0.0f);
		alSource3i(source, AL_AUXILIARY_SEND_FILTER, EnvSlot, 0, AL_FILTER_NULL);
	}
	if(alGetError() != AL_NO_ERROR)
		return NULL;

	alSourcei(source, AL_BUFFER, *static_cast<ALuint*>(sfx.data));
	alSourcePlay(source);
	if(alGetError() != AL_NO_ERROR)
	{
		alSourcei(source, AL_BUFFER, 0);
		alGetError();
		return NULL;
	}

	FreeSfx.pop_back();
	SfxGroup.push_back(source);
	if(!(chanflags&SNDF_NOPAUSE))
	{
		PausableSfx.push_back(source);
		if(SFXPaused)
		{
			alSourcePause(source);
			alGetError();
		}
	}

	FISoundChannel *chan = reuse_chan;
	if(chan != NULL)
		chan->SysChannel = &source;
	else
		chan = S_GetChannel(&source);

	chan->Distance = 0.0f;
	chan->ManualGain = true;

	return chan;
}

FISoundChannel *OpenALSoundRenderer::StartSound3D(SoundHandle sfx, SoundListener *listener, float vol,
	FRolloffInfo *rolloff, float distscale,
	int pitch, int priority, const FVector3 &pos, const FVector3 &vel,
	int /*channum*/, int chanflags, FISoundChannel *reuse_chan)
{
	ALfloat listpos[3];
	alGetListenerfv(AL_POSITION, listpos);
	if(alGetError() != AL_NO_ERROR)
		return NULL;

	listpos[0] =  pos[0] - listpos[0];
	listpos[1] =  pos[1] - listpos[1];
	listpos[2] = -pos[2] - listpos[2];
	float distance = sqrt(listpos[0]*listpos[0] + listpos[1]*listpos[1] +
	                      listpos[2]*listpos[2]) * distscale;

	if(FreeSfx.size() == 0)
	{
		FSoundChan *schan = Channels;
		FSoundChan *lowest = NULL;
		while(schan)
		{
			if(schan->SysChannel != NULL)
			{
				if(!lowest || schan->Priority < lowest->Priority ||
				   (schan->Priority == lowest->Priority &&
				    schan->Distance >= lowest->Distance))
					lowest = schan;
			}
			schan = schan->NextChan;
		}
		if(lowest && (lowest->Priority < priority ||
		   (lowest->Priority == priority && lowest->Distance >= distance)))
			StopChannel(lowest);

		if(FreeSfx.size() == 0)
			return NULL;
	}

	bool manualGain;
	float gain = 1.0f;
	float rolloffFactor;
	ALuint &source = *find(Sources.begin(), Sources.end(), FreeSfx.back());

	alSource3f(source, AL_DIRECTION, 0.0f, 0.0f, 0.0f);
	alSource3f(source, AL_VELOCITY, vel[0], vel[1], -vel[2]);
	alSource3f(source, AL_POSITION, pos[0], pos[1], -pos[2]);
	alSourcei(source, AL_SOURCE_RELATIVE, AL_FALSE);
	alSourcei(source, AL_LOOPING, (chanflags&SNDF_LOOP)?AL_TRUE:AL_FALSE);
	if(rolloff->RolloffType == ROLLOFF_Log)
	{
		if(SrcDistanceModel)
			alSourcei(source, AL_DISTANCE_MODEL, AL_INVERSE_DISTANCE_CLAMPED);
		alSourcef(source, AL_REFERENCE_DISTANCE, rolloff->MinDistance/distscale);
		alSourcef(source, AL_MAX_DISTANCE, (1000.0f+rolloff->MinDistance)/distscale);
		manualGain = false;
		rolloffFactor = rolloff->RolloffFactor;
	}
	else if(SrcDistanceModel)
	{
		if(rolloff->RolloffType == ROLLOFF_Linear)
		{
			alSourcei(source, AL_DISTANCE_MODEL, AL_LINEAR_DISTANCE_CLAMPED);
			alSourcef(source, AL_REFERENCE_DISTANCE, rolloff->MinDistance/distscale);
			alSourcef(source, AL_MAX_DISTANCE, rolloff->MaxDistance/distscale);
			manualGain = false;
			rolloffFactor = 1.0f;
		}
		else
		{
			manualGain = true;
			rolloffFactor = 0.0f;
			gain = GetRolloff(rolloff, distance);
		}
	}
	else
	{
		manualGain = true;
		rolloffFactor = 0.0f;
		gain = GetRolloff(rolloff, distance);
	}
	if(LastWaterLP < 1.0f)
	{
		alSourcef(source, AL_PITCH, pitch/128.0f*PITCH_MULT);
		if(WaterLP)
			alSourcei(source, AL_DIRECT_FILTER, WaterLP);
	}
	else
	{
		alSourcef(source, AL_PITCH, pitch/128.0f);
		if(WaterLP)
			alSourcei(source, AL_DIRECT_FILTER, AL_FILTER_NULL);
	}
	if(EnvSlot)
	{
		alSource3i(source, AL_AUXILIARY_SEND_FILTER, EnvSlot, 0, AL_FILTER_NULL);
		alSourcef(source, AL_ROOM_ROLLOFF_FACTOR, rolloffFactor);
	}
	alSourcef(source, AL_ROLLOFF_FACTOR, rolloffFactor);
	alSourcef(source, AL_GAIN, SfxVolume*gain);
	if(alGetError() != AL_NO_ERROR)
		return NULL;

	alSourcei(source, AL_BUFFER, *static_cast<ALuint*>(sfx.data));
	alSourcePlay(source);
	if(alGetError() != AL_NO_ERROR)
	{
		alSourcei(source, AL_BUFFER, 0);
		alGetError();
		return NULL;
	}

	FreeSfx.pop_back();
	SfxGroup.push_back(source);
	if(!(chanflags&SNDF_NOPAUSE))
	{
		PausableSfx.push_back(source);
		if(SFXPaused)
		{
			alSourcePause(source);
			alGetError();
		}
	}

	FISoundChannel *chan = reuse_chan;
	if(chan != NULL)
		chan->SysChannel = &source;
	else
		chan = S_GetChannel(&source);

	chan->Rolloff = *rolloff;
	chan->DistanceScale = distscale;
	chan->Distance = distance;
	chan->ManualGain = manualGain;

	return chan;
}

void OpenALSoundRenderer::StopChannel(FISoundChannel *chan)
{
	if(chan == NULL || chan->SysChannel == NULL)
		return;

	ALuint source = *static_cast<ALuint*>(chan->SysChannel);
	alSourceStop(source);
	alSourcei(source, AL_BUFFER, 0);
	alGetError();

	std::vector<ALuint>::iterator i = find(PausableSfx.begin(), PausableSfx.end(), source);
	if(i != PausableSfx.end()) PausableSfx.erase(i);
	SfxGroup.erase(find(SfxGroup.begin(), SfxGroup.end(), source));

	FreeSfx.push_back(source);

	chan->SysChannel = NULL;
	S_ChannelEnded(chan);
}

unsigned int OpenALSoundRenderer::GetPosition(FISoundChannel *chan)
{
	if(chan == NULL || chan->SysChannel == NULL)
		return 0;

	ALint pos;
	alGetSourcei(*static_cast<ALuint*>(chan->SysChannel), AL_SAMPLE_OFFSET, &pos);
	if(alGetError() == AL_NO_ERROR)
		return pos;

	return 0;
}


void OpenALSoundRenderer::SetSfxPaused(bool paused, int slot)
{
	int oldslots = SFXPaused;

	if(paused)
		SFXPaused |= 1 << slot;
	else
		SFXPaused &= ~(1 << slot);

	if(oldslots != 0 && SFXPaused == 0)
		alSourcePlayv(PausableSfx.size(), &PausableSfx[0]);
	else if(oldslots == 0 && SFXPaused != 0)
		alSourcePausev(PausableSfx.size(), &PausableSfx[0]);
	alGetError();
}

void OpenALSoundRenderer::SetInactive(bool inactive)
{
}

void OpenALSoundRenderer::Sync(bool sync)
{
}

void OpenALSoundRenderer::UpdateSoundParams3D(SoundListener *listener, FISoundChannel *chan, bool areasound, const FVector3 &pos, const FVector3 &vel)
{
	if(chan == NULL || chan->SysChannel == NULL)
		return;

	ALuint source = *static_cast<ALuint*>(chan->SysChannel);
	alSource3f(source, AL_VELOCITY, vel[0], vel[1], -vel[2]);
	alSource3f(source, AL_POSITION, pos[0], pos[1], -pos[2]);
	alGetError();
}

void OpenALSoundRenderer::UpdateListener(SoundListener *listener)
{
	if(!listener->valid)
		return;

	float angle = listener->angle;
	ALfloat orient[6];
	// forward
	orient[0] = cos(angle);
	orient[1] = 0.0f;
	orient[2] = -sin(angle);
	// up
	orient[3] = 0.0f;
	orient[4] = 1.0f;
	orient[5] = 0.0f;

	alListener3f(AL_POSITION, listener->position.X,
	                          listener->position.Z,
	                         -listener->position.Y);
	alListener3f(AL_VELOCITY, listener->velocity.X,
	                          listener->velocity.Z,
	                         -listener->velocity.Y);
	alListenerfv(AL_ORIENTATION, orient);
	alGetError();


	bool underwater = false;
	const ReverbContainer *env = ForcedEnvironment;
	if(!env)
	{
		underwater = (listener->underwater && *snd_waterhfgain > 0.0f);
		env = listener->Environment;
		if(!env)
			env = DefaultEnvironments[0];
	}

	if(env != PrevEnvironment || env->Modified)
	{
		PrevEnvironment = env;
		DPrintf("Reverb Environment %s\n", env->Name);

		if(EnvSlot)
		{
			ALuint &envReverb = EnvEffects[env->ID];
			bool doLoad = (env->Modified || !envReverb);

			if(!envReverb)
			{
				alGenEffects(1, &envReverb);
				if(alGetError() == AL_NO_ERROR)
				{
					alEffecti(envReverb, AL_EFFECT_TYPE, AL_EFFECT_REVERB);
					if(alGetError() != AL_NO_ERROR)
					{
						alDeleteEffects(1, &envReverb);
						envReverb = 0;
						doLoad = false;
					}
				}
			}

			if(doLoad)
			{
				const_cast<ReverbContainer*>(env)->Modified = false;
#define mB2Gain(x) ((float)pow(10.0, (x)/2000.0))
				alEffectf(envReverb, AL_REVERB_DENSITY,
				          clamp(env->Properties.Density/100.0f, AL_REVERB_MIN_DENSITY, AL_REVERB_MAX_DENSITY));
				alEffectf(envReverb, AL_REVERB_DIFFUSION,
				          clamp(env->Properties.Diffusion/100.0f, AL_REVERB_MIN_DIFFUSION, AL_REVERB_MAX_DIFFUSION));
				alEffectf(envReverb, AL_REVERB_GAIN,
				          clamp(mB2Gain(env->Properties.Room), AL_REVERB_MIN_GAIN, AL_REVERB_MAX_GAIN));
				alEffectf(envReverb, AL_REVERB_GAINHF,
				          clamp(mB2Gain(env->Properties.RoomHF), AL_REVERB_MIN_GAINHF, AL_REVERB_MAX_GAINHF));
				alEffectf(envReverb, AL_REVERB_DECAY_TIME,
				          clamp(env->Properties.DecayTime, AL_REVERB_MIN_DECAY_TIME, AL_REVERB_MAX_DECAY_TIME));
				alEffectf(envReverb, AL_REVERB_DECAY_HFRATIO,
				          clamp(env->Properties.DecayHFRatio, AL_REVERB_MIN_DECAY_HFRATIO, AL_REVERB_MAX_DECAY_HFRATIO));
				alEffectf(envReverb, AL_REVERB_REFLECTIONS_GAIN,
				          clamp(mB2Gain(env->Properties.Reflections), AL_REVERB_MIN_REFLECTIONS_GAIN, AL_REVERB_MAX_REFLECTIONS_GAIN));
				alEffectf(envReverb, AL_REVERB_REFLECTIONS_DELAY,
				          clamp(env->Properties.ReflectionsDelay, AL_REVERB_MIN_REFLECTIONS_DELAY, AL_REVERB_MAX_REFLECTIONS_DELAY));
				alEffectf(envReverb, AL_REVERB_LATE_REVERB_GAIN,
				          clamp(mB2Gain(env->Properties.Reverb), AL_REVERB_MIN_LATE_REVERB_GAIN, AL_REVERB_MAX_LATE_REVERB_GAIN));
				alEffectf(envReverb, AL_REVERB_LATE_REVERB_DELAY,
				          clamp(env->Properties.ReverbDelay, AL_REVERB_MIN_LATE_REVERB_DELAY, AL_REVERB_MAX_LATE_REVERB_DELAY));
				alEffectf(envReverb, AL_REVERB_AIR_ABSORPTION_GAINHF,
				          clamp(mB2Gain(env->Properties.AirAbsorptionHF), AL_REVERB_MIN_AIR_ABSORPTION_GAINHF, AL_REVERB_MAX_AIR_ABSORPTION_GAINHF));
				alEffectf(envReverb, AL_REVERB_ROOM_ROLLOFF_FACTOR,
				          clamp(env->Properties.RoomRolloffFactor, AL_REVERB_MIN_ROOM_ROLLOFF_FACTOR, AL_REVERB_MAX_ROOM_ROLLOFF_FACTOR));
				alEffecti(envReverb, AL_REVERB_DECAY_HFLIMIT,
				          (env->Properties.Flags&REVERB_FLAGS_DECAYHFLIMIT)?AL_TRUE:AL_FALSE);
#undef mB2Gain
			}

			alAuxiliaryEffectSloti(EnvSlot, AL_EFFECTSLOT_EFFECT, envReverb);
			alGetError();
		}
	}

	// NOTE: Moving into and out of water (and changing the gainhf) will undo
	// pitch variations on the sounds.
	if(underwater || env->SoftwareWater)
	{
		if (LastWaterLP != *snd_waterhfgain)
		{
			LastWaterLP = *snd_waterhfgain;

			foreach(ALuint, i, SfxGroup)
				alSourcef(*i, AL_PITCH, PITCH_MULT);
			alGetError();

			if(WaterLP != 0)
			{
				alEffectf(WaterLP, AL_LOWPASS_GAIN, 1.0f);
				alEffectf(WaterLP, AL_LOWPASS_GAINHF, LastWaterLP);
				if(alGetError() != AL_NO_ERROR)
				{
					// Apply the updated filter on the sources
					foreach(ALuint, i, SfxGroup)
						alSourcei(*i, AL_DIRECT_FILTER, WaterLP);
					alGetError();
				}
			}
		}
	}
	else
	{
		if (LastWaterLP < 1.0f)
		{
			LastWaterLP = 1.0f;

			foreach(ALuint, i, SfxGroup)
				alSourcef(*i, AL_PITCH, 1.0f);
			alGetError();

			if(WaterLP != 0)
			{
				foreach(ALuint, i, SfxGroup)
					alSourcei(*i, AL_DIRECT_FILTER, AL_FILTER_NULL);
				alGetError();
			}
		}
	}
}

void OpenALSoundRenderer::UpdateSounds()
{
	// Not all sources can use the distance models provided by OpenAL. For the
	// ones that can't, apply the calculated attenuation as the source gain to
	// the 3D source.
	ALfloat listenerpos[3] = { 0.0f, 0.0f, 0.0f };
	alGetListenerfv(AL_POSITION, listenerpos);
	alGetError();

	FSoundChan *schan = Channels;
	while(schan)
	{
		if(schan->SysChannel != NULL)
		{
			ALuint source = *static_cast<ALuint*>(schan->SysChannel);
			ALfloat sourcepos[3];
			ALint state;
			ALint local;

			alGetSourcei(source, AL_SOURCE_STATE, &state);
			alGetSourcei(source, AL_SOURCE_RELATIVE, &local);
			alGetSourcefv(source, AL_POSITION, sourcepos);
			if(alGetError() == AL_NO_ERROR)
			{
				if(state != AL_PLAYING && state != AL_PAUSED)
				{
					// If this channel is stopped, release it
					StopChannel(schan);
				}
				else if(local == AL_FALSE)
				{
					// Sources that are head-relative (local) don't need
					// attenuation
					float gain;

					sourcepos[0] -= listenerpos[0];
					sourcepos[1] -= listenerpos[1];
					sourcepos[2] -= listenerpos[2];
					schan->Distance = sqrt(sourcepos[0]*sourcepos[0] + sourcepos[1]*sourcepos[1] +
					                       sourcepos[2]*sourcepos[2]) * schan->DistanceScale;

					if(schan->ManualGain)
					{
						float gain = GetRolloff(&schan->Rolloff, schan->Distance);
						alSourcef(source, AL_GAIN, SfxVolume*gain);
						alGetError();
					}
				}
			}
		}

		schan = schan->NextChan;
	}

	foreach(OpenALSoundStream*, i, Streams)
		(*i)->Poll();
}

bool OpenALSoundRenderer::IsValid ()
{
	return Context != NULL;
}

void OpenALSoundRenderer::PrintStatus ()
{
	ALCdevice *Device = alcGetContextsDevice(Context);
	Printf("Output: "TEXTCOLOR_ORANGE"%s\n", alcGetString(Device, ALC_DEVICE_SPECIFIER));
	Printf("Driver caps:");
	if(alcIsExtensionPresent(Device, "ALC_EXT_EFX"))
		Printf(TEXTCOLOR_GREEN" EFX");
	if(alIsExtensionPresent("AL_EXT_FLOAT32"))
		Printf(","TEXTCOLOR_GREEN" Float formats");
	if(alIsExtensionPresent("AL_EXT_MCFORMATS"))
		Printf(","TEXTCOLOR_GREEN" Multi-channel Formats");
	Printf("\n");
	alcGetError(Device);

	Printf("Max sound sources: "TEXTCOLOR_BLUE"%d\n", Sources.size());
}

FString OpenALSoundRenderer::GatherStats()
{
	ALCdevice *Device = alcGetContextsDevice(Context);
	ALCint update = 0;
	alcGetIntegerv(Device, ALC_REFRESH, 1, &update);
	alcGetError(Device);

	ALuint total = Sources.size();
	ALuint used = SfxGroup.size()+Streams.size();
	ALuint unused = FreeSfx.size();
	FString out;
	out.Format("%u sources ("TEXTCOLOR_YELLOW"%u"TEXTCOLOR_NORMAL" active, "TEXTCOLOR_YELLOW"%u"TEXTCOLOR_NORMAL" free), "
	           "Update interval: "TEXTCOLOR_YELLOW"%d"TEXTCOLOR_NORMAL"ms",
	           total, used, unused, 1000/update);
	return out;
}

void OpenALSoundRenderer::PrintDriversList ()
{
	ALCsizei count = 0;
	const ALCchar **drivers = alureGetDeviceNames(ALC_TRUE, &count);
	if(drivers == NULL)
	{
		Printf(TEXTCOLOR_YELLOW"Failed to retieve device list: %s\n", alureGetErrorString());
		return;
	}

	for(int i = 0;i < count;i++)
		Printf("%d. %s\n", i, drivers[i]);
	alureFreeDeviceNames(drivers);
}

#endif // NO_OPENAL
